Remove["ObjectMath`Notebook`*"];
Remove["ObjectMath`Notebook`*`*"];

BeginPackage["ObjectMath`Notebook`"];

Needs["ObjectMath`", "ObjectMath.m"];

viewObject::usage = "viewObject[\"package\", \"object\"] To be written...";

viewComponent::usage = 
  "viewComponent[\"package\", \"object\", \"component\"] To be written...";

gotoObject::usage = "gotoObject[\"package\", \"object\"] To be written...";

gotoComponent::usage = 
  "gotoComponent[\"package\", \"object\", \"component\"] To be written...";

setObjectCoordinates::usage = 
  "setObjectCoordinates[\"package\", \"object\", \"coordinates\"] To be \
written...";

setComponentCoordinates::usage = 
  "setComponentCoordinates[\"package\", \"object\", \"component\", \
\"coordinates\"] To be written...";

register::usage = 
  "register[\"package\"] Register a package.  Must be done from the current \
notebook.  This function also tells the browser that the package exists.";

unregister::usage = "unregister[\"package\"] Unregister a package.";

fetchModel::usage = "fetchModel[\"package\"] Refresh.  Does a storeModel.";

storeModel::usage = "storeModel[\"package\"] Store the model.";


fModel[] := 
  ObjectMath`Notebook`Private`pfetchModel["Testing"];


Begin["`Private`"];

(*
 * 
 * Definition of module variables.
 *
 *)

browserPackages = {};

(*
 *
 * Definitions of public functions.
 *
 *)

defaultModel = "OBJ $ $ ROOT 0 0 0 0\nOBJ NONEXISTANT $ CLASS 0 0 0 0\n";


fetchModel[package_] := (
  loadCoordinates[package];
  storeModel[package];
  If[ValueQ[storedModel[package]],
    storedModel[package],
    defaultModel]
)

storeModel[package_] :=
  Module[{info, res},
    Begin["ObjectMath`CollectBogusSymbolDefinitions`"];
    info = getBrowserInfoPackage[package];
    If[MatchQ[info, _Integer], res = defaultModel,
       res = StringJoin[rootObjectString[package],
                        browserObjectString[package, info],    
                        browserInhString[info],
                        "ORI LR\n"]];
    End[];
    Remove["ObjectMath`CollectBogusSymbolDefinitions`*"];
    storedModel[package] = res;
  ];

register[package_] := (
  browserPackages = Append[browserPackages, package];
  browserNotebooks[package]= SelectedNotebook[];
(* The following does not work.  It ought to work.

Fails:   FrontEnd`FrontEndExecute[
           ObjectMath`Notebook`Private`om$$ToBrowser[register Testing]]

Works:   EnterTextPacket[\012FrontEndExecute[om$$ToBrowser["register Testing"]]]
*)
(*  FrontEndExecute[om$$ToBrowser[StringJoin["register ", package]]]; *)
)

unregister[package_] := (
  DeleteCases[browserPackages, package];
  browserNotebooks[package] = 0;
)

(* Did goto class3 and after that class2 and Mathematica crashed *)

gotoObject[package_, object_] := (
(*  NotebookFind[browserNotebooks[package], object, All, CellTags]; *)
  NotebookLocate[object];
  0
)

viewObject[package_, object_] := (
  Module[
    {exp, viewnb, title},
    NotebookLocate[object];
    exp = NotebookRead[browserNotebooks[package]];
    title = StringJoin["View Window (", package, ":", object, ")"];
    viewnb = NotebookCreate[];
    SetOptions[viewnb, WindowTitle -> title, WindowSize -> {400., 200.}];
    NotebookWrite[viewnb, exp];
    SetOptions[viewnb, Deletable->False, Editable->False];
    0
  ]
)

gotoComponent[package_, object_, component_] := (
(*  NotebookFind[browserNotebooks[package], object, All, CellTags]; *)
  NotebookLocate[object];
  NotebookFind[browserNotebooks[package], component, Previous];
  NotebookFind[browserNotebooks[package], component];
(*  NotebookLocate[StringJoin[object, ":", component]]; *)
  0
)

(*
 * 
 * Private functions.
 *
 *)


(* MakeExpression introduces bogus symbols *)

pfetchModel[package_] := 
  Module[{info, res},
      res = 17;
(*    Begin["ObjectMath`CollectBogusSymbolDefinitions`"]; *)
         info = getBrowserInfoPackage[package];
(*         If[MatchQ[info, _Integer], res = info,
       res = StringJoin[browserObjectString[package, info],    
                        browserInhString[info],
                        "ORI LR\n"]];*)
(*    End[];
    Remove["ObjectMath`CollectBogusSymbolDefinitions`*"]; *)
    res
  ];


(* This is not used.
tokenize[Notebook[x_, y___]] := Map[tokenize, x];
tokenize[Cell[x_, "Input"]] := tokenize[x];
tokenize[BoxData[x___]] := tokenize[x];
tokenize[RowBox[x___]] := tokenize[x];
(* NotebookTokenize[{"class", x___}] := getc[x]; *)
tokenize[{x___, RowBox[y___], z___}] := tokenize[Flatten[{x, y, z}]];
tokenize[x___] := x;
*)


makeExp[Cell[CellGroupData[x___], y___]] :=  {};
makeExp[Cell[x_, y___]] :=  makeExp[x];
makeExp[TextData[x_]] :=  MakeExpression[x];
makeExp[BoxData[x_]] :=  {};
makeExp[x_] :=  {};


getClassX[HoldComplete[Class[x_,y___];z___]] :=  SymbolName[x[[0]]];
getClassX[HoldComplete[Class[x_;y___];z___]] :=  SymbolName[x[[0]]];
(* getClassX[ErrorBox[x___]] := "Syntax error in cell"; *)
getClassX[x___] = {};

getClass[HoldComplete[Class[x_Symbol,y___];z___]] := SymbolName[x];
getClass[HoldComplete[Class[x_Symbol;y___];z___]] := SymbolName[x];
getClass[HoldComplete[Class[x_,y___];z___]] := SymbolName[x[[0]]];
getClass[HoldComplete[Class[x_;y___];z___]] := SymbolName[x[[0]]];
(* getClass[ErrorBox[x___]] := "Syntax error in cell"; *)
getClass[x___] = {};

getInstance[HoldComplete[Instance[x_Symbol,y___];z___]] := SymbolName[x];
getInstance[HoldComplete[Instance[x_Symbol;y___];z___]] := SymbolName[x];
getInstance[HoldComplete[Instance[x_,y___];z___]] := SymbolName[x[[0]]];
getInstance[HoldComplete[Instance[x_;y___];z___]] := SymbolName[x[[0]]];
(* getInstance[ErrorBox[x___]] := "Syntax error in cell"; *)
getInstance[x___] = {};

getInherit[HoldComplete[Class[x_,y___];z___]] := 
  getInherit[HoldComplete[y]];
getInherit[HoldComplete[Class[x_;y___];z___]] := 
  getInherit[HoldComplete[y]];
getInherit[HoldComplete[Instance[x_,y___];z___]] :=  
  getInherit[HoldComplete[y]];
getInherit[HoldComplete[Instance[x_;y___];z___]] :=  
  getInherit[HoldComplete[y]];
getInherit[HoldComplete[Inherit[x_],y___]] := 
  Join[{SymbolName[x[[0]]]}, getInherit[HoldComplete[y]]];
getInherit[HoldComplete[Inherit[x_];y___]] := 
  Join[{SymbolName[x[[0]]]}, getInherit[HoldComplete[y]]];

(* Extends instead of Inherits *)
getInherit[HoldComplete[Extends[x_Symbol],y___]] := 
  Join[{SymbolName[x]}, getInherit[HoldComplete[y]]];
getInherit[HoldComplete[Extends[x_Symbol];y___]] := 
  Join[{SymbolName[x]}, getInherit[HoldComplete[y]]];
getInherit[HoldComplete[Extends[x_],y___]] := 
  Join[{SymbolName[x[[0]]]}, getInherit[HoldComplete[y]]];
getInherit[HoldComplete[Extends[x_];y___]] := 
  Join[{SymbolName[x[[0]]]}, getInherit[HoldComplete[y]]];

getInherit[x___] := {};


getComponents[HoldComplete[Class[x_,y___];z___]] := 
  getComponents[HoldComplete[y]];
getComponents[HoldComplete[Class[x_;y___];z___]] := 
  getComponents[HoldComplete[y]];
getComponents[HoldComplete[Instance[x_,y___];z___]] := 
  getComponents[HoldComplete[y]];
getComponents[HoldComplete[Instance[x_;y___];z___]] := 
  getComponents[HoldComplete[y]];
getComponents[HoldComplete[x___, Components[y___], z___]] :=
  getComponents[HoldComplete[y]];
getComponents[HoldComplete[x___; Components[y___]; z___]] :=
  getComponents[HoldComplete[y]];


getComponents[HoldComplete[Component[x_, y___], z___]] :=
  Join[{SymbolName[x[[0]]], getInherit[HoldComplete[y]]}, 
       getComponents[HoldComplete[z]]];
getComponents[HoldComplete[Component[x_; y___], z___]] :=
  Join[{SymbolName[x[[0]]], getInherit[HoldComplete[y]]}, 
       getComponents[HoldComplete[z]]];
getComponents[HoldComplete[Component[x_, y___]; z___]] :=
  Join[{SymbolName[x[[0]]], getInherit[HoldComplete[y]]}, 
       getComponents[HoldComplete[z]]];
getComponents[HoldComplete[Component[x_; y___]; z___]] :=
  Join[{SymbolName[x[[0]]], getInherit[HoldComplete[y]]}, 
       getComponents[HoldComplete[z]]];

(*
getComponents[HoldComplete[Component[x_, y___], z___]] := "commacomp";
getComponents[HoldComplete[Component[x_; y___], z___]] := "semicomp";
*)
getComponents[x___] := {};

(* Return a list {name, type, component of, list of inh, coord}
        If coord is {} the placement is the default placement.
                        *)

getBrowserInfo[nbe_] :=
        Module[{cl, in, inh, comp, nodes, rels, i, name, type, compof, Inherit},
                nodes = {};
                rels = {};
                cl = getClass /@ makeExp /@ nbe[[1]];
                in = getInstance /@ makeExp /@ nbe[[1]];
                inh = getInherit /@ makeExp /@ nbe[[1]];
                comp = getComponents /@ makeExp /@ nbe[[1]];
                For[i=1, i<=Length[cl], i++,
                  type = "";
                        If [cl[[i]] =!= {}, name= cl[[i]]; type ="CLASS";];
                        If [in[[i]] =!= {}, name= in[[i]]; type = "INSTANCE";];
                        If[type =!= "",
       nodes = Append[nodes, {name, type, "$", inh[[i]]}];
                        If [comp[[i]] =!= {},
               For[j=1,j<=Length [comp[[i]]],j = j + 2,
                       nodes = Append[nodes, 
                                    {comp[[i,j]], "PART",
                                             name,comp[[i, j+1]]}];];];];
                        ]; 
                        (* Print[i, in[[i]], cl[[i]] == {}, tmp]; *)
                        nodes
        ];


getBrowserInfotemp[nbe_] :=
        Module[{cl, in, inh, comp, nodes, rels, i, name, type, compof, Inherit},
                nodes = {};
                rels = {};
	        cl = getClass /@ makeExp /@ nbe[[1]];
                in = getInstance /@ makeExp /@ nbe[[1]];
                inh = getInherit /@ makeExp /@ nbe[[1]];
                comp = getComponents /@ makeExp /@ nbe[[1]];
      {cl, in, inh, comp, (makeExp /@ nbe[[1]])}
        ];



getBrowserInfoPackage[package_] :=
  Module[{nb, nbe, res},
    nb = browserNotebooks[package];
    nbe = "No notebook";
    res = 1;
    If[MatchQ[nb, _NotebookObject],
        (nbe = NotebookGet[nb];
        If[MatchQ[nbe, _Notebook], res = getBrowserInfo[nbe], res = 1];)
    ];
    If[res =!= 1, tagNotebook[nb, nbe]];
    res
  ];


testgetBrowserInfoPackage[package_] :=
  Module[{nb, nbe, res},
    nb = browserNotebooks[package];
    nbe = "No notebook";
    res = 1;
    If[MatchQ[nb, _NotebookObject],
        (nbe = NotebookGet[nb];
        If[MatchQ[nbe, _Notebook], res = getBrowserInfotemp[nbe], res = 1];)
    ];
(*    If[res =!= 1, tagNotebook[nb, nbe]]; *)
    res
  ];



(*
 *
 * Tag the notebook
 *
 *)

tagNotebook[nb_, nbe_] := Module[{nbe2},
  nbe2 = nbe //. {Cell[TextData[x_?totagQ], "Input", y___] :>
           Cell[TextData[x], "Input", gettag[x]]};
(*  Print[nbe2]; *)
  NotebookPut[nbe2, nb];
  ];

(* ToHeldExpression is obsolete (use ToExpression[expr, form, Hold] *)
(* must erase the tag if it is defined *)

totagQ[x_String] := totagQ[ToHeldExpression[x]];
totagQ[Hold[___,Class[_, ___]]] := True;
totagQ[Hold[___,Class[_, ___];]] := True;
totagQ[Hold[___,Class[_, ___];];] := True;
totagQ[Hold[___,Class[_; ___]]] := True;
totagQ[Hold[___,Class[_; ___];]] := True;
totagQ[Hold[___,Class[_; ___];];] := True;
totagQ[Hold[___,Instance[_, ___]]] := True;
totagQ[Hold[___,Instance[_, ___];]] := True;
totagQ[Hold[___,Instance[_, ___];];] := True;
totagQ[Hold[___,Instance[_; ___]]] := True;
totagQ[Hold[___,Instance[_; ___];]] := True;
totagQ[Hold[___,Instance[_; ___];];] := True;
totagQ[x_] := False;

gettag[x_String] := gettag[ToHeldExpression[x]];

gettag[Hold[___,Class[x_, y___]]] := torule[x];
gettag[Hold[___,Class[x_, y___];]] := torule[x];
gettag[Hold[___,Class[x_, y___];];] := torule[x];
gettag[Hold[___,Class[x_; y___]]] := torule[x];;
gettag[Hold[___,Class[x_; y___];]] := torule[x];
gettag[Hold[___,Class[x_; y___];];] := torule[x];
gettag[Hold[___,Instance[x_, y___]]] := CellTags -> {ToString[Head[x]]};
gettag[Hold[___,Instance[x_, y___];]] := CellTags -> {ToString[Head[x]]};
gettag[Hold[___,Instance[x_, y___];];] := CellTags -> {ToString[Head[x]]};
gettag[Hold[___,Instance[x_; y___]]] := CellTags -> {ToString[Head[x]]};
gettag[Hold[___,Instance[x_; y___];]] := CellTags -> {ToString[Head[x]]};
gettag[Hold[___,Instance[x_; y___];];] := CellTags -> {ToString[Head[x]]};

gettag[x_] := "";

torule[x_Symbol] := CellTags -> {ToString[x]};
torule[x_] := CellTags -> {ToString[Head[x]]};


(*
 *
 * Get the string represenation.
 *
 *)

rootObjectString[package_] := 
  StringJoin["OBJ $ $ ROOT ", getCoordinates[package, "$", "$"], "\n"];

browserInhString[{x_List, y___}] := 
  StringJoin[browserInhString[x], browserInhString[{y}]];
browserInhString[{}] := "";
browserInhString[{name_String, type_String, compof_String, inh_List}] := 
  StringJoin[
    Map[StringJoin["INH ", name, " ", compof, " ", #, "\n"]&, inh]];
(*  {TYPES, name[[0]], type[[0]], compof[[0]], inh[[0]]}; *)


browserObjectString[package_, {x_List, y___}] :=
  StringJoin[browserObjectString[package, x], 
             browserObjectString[package, {y}]];
browserObjectString[package_, {}] := "";
browserObjectString [package_, 
                     {name_String, type_String, compof_String, inh_List}] := 
  StringJoin["OBJ ", name, " ", compof, " ", type, " ", 
             getCoordinates[package, name, compof], "\n"]

(*
 * Coordinate stuff
 *
 *)

setObjectCoordinates[package_, object_, coord_] := (
  setCoordinates[package, object, "$", coord]
)

setComponentCoordinates[package_, object_, component_, coord_] := (
  setCoordinates[package, object, component, coord]
)

(*
 * Store the value of browserCoordinates[package] in the option ConversionRules 
 * option in the BeginPackage cell.
 *
 *)

storeCoordinates[package_] := (
  Module[{nb},
    nb = browserNotebooks[package];
    NotebookFind[nb, StringJoin["BeginPackage[\"", package, "`\"]"]];
    SelectionMove[nb, All, Cell];
    SetOptions[NotebookSelection[nb], 
               ConversionRules -> {"coordinates"->browserCoordinates[package]}]
  ]
)

(*
 * Assign the value stored in the ConversionRules option in the BeginPackage
 * to the variable browserCoordinates[package].
 *
 *)

loadCoordinates[package_] := (
  Module[{nb, cr},
    nb = browserNotebooks[package];
    NotebookFind[nb, StringJoin["BeginPackage[\"", package, "`\"]"]];
    SelectionMove[nb, All, Cell];
    cr = ConversionRules /. Options[NotebookSelection[nb], ConversionRules];
    If[cr === None,
      browserCoordinates[package] = {},
      browserCoordinates[package] = "coordinates" /. cr];
  ]
)

setCoordinates[package_, name_, compof_, coord_] := 
  Module[{c, head, tail},
    If[Not[ValueQ[browserCoordinates[package]]], browserCoordinates[package] = {}];
    If[MatchQ[browserCoordinates[package], {head___, {name, compof, c_},tail___}],
      browserCoordinates[package] = browserCoordinates[package] /. {name, compof, c_}->{name, compof, coord},
      browserCoordinates[package] = Append[browserCoordinates[package], 
                                           {name, compof, coord}]
    ];
    storeCoordinates[package];
    0
  ]    



getCoordinates[package_, name_, compof_] := (
  If[Not[ValueQ[browserCoordinates[package]]], browserCoordinates[package] = {}];
  getC[name, compof, browserCoordinates[package]]
)

getC[name_, compof_, {head_, tail___}] := 
  If[MatchQ[head, {name, compof, c_}], head[[3]], getC[name, compof, {tail}]];
getC[name_, compof_, {}] := "0 0 0 0";

(*
coordinatestostring[x_] := 
    StringJoin[
      Map[Function[y, StringJoin[ToString[y], " "]], 
          Flatten[x]]]
*)

End[]; (* Private *)

EndPackage[];



