! MathLinkInterface.f90

! $Id$

! This file provides an interface between the MathCode generated
! Fortran90 code and MathLink. It calls functions in the file
! MathLinkWrapper.c which contains the actual calls to functions
! like MLGetInteger. It also contains the PROGRAM part of the
! MathCode application which simply passes the command line to
! MLIF_MAIN in MathLinkWrapper.c and further on to MLMain.

! The reason that these functions are included in a module is
! the fact that many of them have pointers as dummy arguments
! (Fortan90 lingo for formal parameters). This means that they
! have to have an explicit interface. The module construct is
! easier than having a bunch of global functions and including
! interfaces in the code later.

! -----------------------------------------------------------------
!  Start of module

module MathLinkInterface
	use MathCodePrecision
	implicit none
contains

! -----------------------------------------------------------------
!  Transfer simple variables

subroutine putNull()
       	call SENDNULL()
end subroutine putNull

subroutine getint(i)
	integer(mc_int) i
	call RECEIVEINT(i)
end subroutine getint

subroutine getstr(s,s_i)
        character(len=mc_strlen) s
        integer(mc_int) s_i
        integer(mc_int) i
	call RECEIVESTR(s,s_i)
        do i=s_i+1,mc_strlen
          s(i:i)=' '
        end do
end subroutine getstr

subroutine putstr(s)
        character(len=mc_strlen) s
	call SENDSTR(s)
end subroutine putstr

subroutine putint(i)
	integer(mc_int) i
	call SENDINT(i)
end subroutine putint

subroutine getdouble(d)
	real(mc_real) d
	call RECEIVEDOUBLE(d)
end subroutine getdouble

subroutine putdouble(d)
	real(mc_real) d
	call SENDDOUBLE(d)
end subroutine putdouble

! TODO: Write get and put for strings and vectors of strings

! -----------------------------------------------------------------
!  Transfer vectors

subroutine getNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:)
	integer(mc_int) remote_depth
	call RECEIVENINT(remote_depth)
	allocate(remote_data_int(remote_depth))
	call COPYNINT(remote_data_int)
end subroutine getNint

subroutine putNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:)
	integer(mc_int) remote_depth
	remote_depth = size(remote_data_int,1)
	call SENDNINT(remote_data_int, remote_depth)
end subroutine putNint

subroutine getNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:)
	integer(mc_int) remote_depth
	call RECEIVENDOUBLE(remote_depth)
	allocate(remote_data_double(remote_depth))
	call COPYNDOUBLE(remote_data_double)
end subroutine getNdouble

subroutine putNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:)
	integer(mc_int) remote_depth
	remote_depth = size(remote_data_double,1)
	call SENDNDOUBLE(remote_data_double, remote_depth)    
end subroutine putNdouble

! -----------------------------------------------------------------
!  Transfer 2D arrays

subroutine getNNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:,:)
	integer(mc_int) remote_dims(2)
	call RECEIVENNINT(remote_dims)
	allocate(remote_data_int(remote_dims(2), remote_dims(1)))
	call COPYNNINT(remote_data_int)
end subroutine getNNint

subroutine putNNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:,:)
	integer(mc_int) remote_dims(2)
	!remote_dims = shape(remote_data_int)
	remote_dims(2) = size(remote_data_int,1)
	remote_dims(1) = size(remote_data_int,2)
	call SENDNNINT(remote_data_int, remote_dims)
end subroutine putNNint

subroutine getNNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:,:)
	integer(mc_int) remote_dims(2)
	call RECEIVENNDOUBLE(remote_dims)
	allocate(remote_data_double(remote_dims(2), remote_dims(1)))
	call COPYNNDOUBLE(remote_data_double)
end subroutine getNNdouble

subroutine putNNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:,:)
	integer(mc_int) remote_dims(2)
	!remote_dims = shape(remote_data_double)
	remote_dims(2) = size(remote_data_double,1)
	remote_dims(1) = size(remote_data_double,2)
	call SENDNNDOUBLE(remote_data_double, remote_dims)
end subroutine putNNdouble

! -----------------------------------------------------------------
!  Transfer 3D arrays

subroutine getNNNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:,:,:)
	integer(mc_int) remote_dims(3)
	call RECEIVENNNINT(remote_dims)
	allocate(remote_data_int(remote_dims(3), remote_dims(2), remote_dims(1)))
	call COPYNNNINT(remote_data_int)
end subroutine getNNNint

subroutine putNNNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:,:,:)
	integer(mc_int) remote_dims(3)
	!remote_dims = shape(remote_data_int)
	remote_dims(3) = size(remote_data_int,1)
	remote_dims(2) = size(remote_data_int,2)
	remote_dims(1) = size(remote_data_int,3)
	call SENDNNNINT(remote_data_int, remote_dims)
end subroutine putNNNint

subroutine getNNNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:,:,:)
	integer(mc_int) remote_dims(3)
	call RECEIVENNNDOUBLE(remote_dims)
	allocate(remote_data_double(remote_dims(3), remote_dims(2), remote_dims(1)))
	call COPYNNNDOUBLE(remote_data_double)
end subroutine getNNNdouble

subroutine putNNNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:,:,:)
	integer(mc_int) remote_dims(3)
	!remote_dims = shape(remote_data_double)
	remote_dims(3) = size(remote_data_double,1)
	remote_dims(2) = size(remote_data_double,2)
	remote_dims(1) = size(remote_data_double,3)
	call SENDNNNDOUBLE(remote_data_double, remote_dims)
end subroutine putNNNdouble

! -----------------------------------------------------------------
!  Transfer 4D arrays

subroutine getNNNNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:,:,:,:)
	integer(mc_int) remote_dims(4)
	call RECEIVENNNNINT(remote_dims)
	allocate(remote_data_int(remote_dims(4), remote_dims(3), remote_dims(2), remote_dims(1)))
	call COPYNNNNINT(remote_data_int)
end subroutine getNNNNint

subroutine putNNNNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:,:,:,:)
	integer(mc_int) remote_dims(4)
	!remote_dims = shape(remote_data_int)
	remote_dims(4) = size(remote_data_int,1)
	remote_dims(3) = size(remote_data_int,2)
	remote_dims(2) = size(remote_data_int,3)
	remote_dims(1) = size(remote_data_int,4)
	call SENDNNNNINT(remote_data_int, remote_dims)
end subroutine putNNNNint

subroutine getNNNNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:,:,:,:)
	integer(mc_int) remote_dims(4)
	call RECEIVENNNNDOUBLE(remote_dims)
!     Using these commands is not compatible with ABSOFT!	
!	!DEC$ IF DEFINED (IFDEBUG)
!	print *, "Got 4D real array with C dimensions ", remote_dims
!	call flush(6)
!	!DEC$ ENDIF
	allocate(remote_data_double(remote_dims(4), remote_dims(3), remote_dims(2), remote_dims(1)))
	call COPYNNNNDOUBLE(remote_data_double)
end subroutine getNNNNdouble

subroutine putNNNNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:,:,:,:)
	integer(mc_int) remote_dims(4)
	!remote_dims = shape(remote_data_double)
	remote_dims(4) = size(remote_data_double,1)
	remote_dims(3) = size(remote_data_double,2)
	remote_dims(2) = size(remote_data_double,3)
	remote_dims(1) = size(remote_data_double,4)
	call SENDNNNNDOUBLE(remote_data_double, remote_dims)
end subroutine putNNNNdouble

! -----------------------------------------------------------------
!  Transfer simple variables (callback)

subroutine cbgetnull()
	call CBRECEIVENULL()
end subroutine cbgetnull

subroutine cbgetint(i)
	integer(mc_int) i
	call CBRECEIVEINT(i)
end subroutine cbgetint

subroutine cbgetdouble(d)
	real(mc_real) d
	call CBRECEIVEDOUBLE(d)
end subroutine cbgetdouble

! TODO: Write cbget for strings and vectors of strings

! -----------------------------------------------------------------
!  Transfer vectors (callback)

subroutine cbgetNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:)
	integer(mc_int) remote_depth
	call CBRECEIVENINT(remote_depth)
	allocate(remote_data_int(remote_depth))
	call CBCOPYNINT(remote_data_int)
end subroutine cbgetNint

subroutine cbgetNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:)
	integer(mc_int) remote_depth
	call CBRECEIVENDOUBLE(remote_depth)
	allocate(remote_data_double(remote_depth))
	call CBCOPYNDOUBLE(remote_data_double)
end subroutine cbgetNdouble

! -----------------------------------------------------------------
!  Transfer 2D arrays (callback)

subroutine cbgetNNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:,:)
	integer(mc_int) remote_dims(2)
	call CBRECEIVENNINT(remote_dims)
	allocate(remote_data_int(remote_dims(2), remote_dims(1)))
	call CBCOPYNNINT(remote_data_int)
end subroutine cbgetNNint

subroutine cbgetNNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:,:)
	integer(mc_int) remote_dims(2)
	call CBRECEIVENNDOUBLE(remote_dims)
	allocate(remote_data_double(remote_dims(2), remote_dims(1)))
	call CBCOPYNNDOUBLE(remote_data_double)
end subroutine cbgetNNdouble

! -----------------------------------------------------------------
!  Transfer 3D arrays (callback)

subroutine cbgetNNNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:,:,:)
	integer(mc_int) remote_dims(3)
	call CBRECEIVENNNINT(remote_dims)
	allocate(remote_data_int(remote_dims(3), remote_dims(2), remote_dims(1)))
	call CBCOPYNNNINT(remote_data_int)
end subroutine cbgetNNNint

subroutine cbgetNNNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:,:,:)
	integer(mc_int) remote_dims(3)
	call CBRECEIVENNNDOUBLE(remote_dims)
	allocate(remote_data_double(remote_dims(3), remote_dims(2), remote_dims(1)))
	call CBCOPYNNNDOUBLE(remote_data_double)
end subroutine cbgetNNNdouble

! -----------------------------------------------------------------
!  Transfer 4D arrays (callback)

subroutine cbgetNNNNint(remote_data_int)
	integer(mc_int), pointer :: remote_data_int(:,:,:,:)
	integer(mc_int) remote_dims(4)
	call CBRECEIVENNNNINT(remote_dims)
	allocate(remote_data_int(remote_dims(4), remote_dims(3), remote_dims(2), remote_dims(1)))
	call CBCOPYNNNNINT(remote_data_int)
end subroutine cbgetNNNNint

subroutine cbgetNNNNdouble(remote_data_double)
	real(mc_real), pointer :: remote_data_double(:,:,:,:)
	integer(mc_int) remote_dims(4)
	call CBRECEIVENNNNDOUBLE(remote_dims)
	allocate(remote_data_double(remote_dims(4), remote_dims(3), remote_dims(2), remote_dims(1)))
	call CBCOPYNNNNDOUBLE(remote_data_double)
end subroutine cbgetNNNNdouble

!
! -----------------------------------------------------------------
!  End of module

end module MathLinkInterface
